#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include "tscp.h"

FQSender::FQSender(Network *n, int p, int pktSize, float think)
{
    net = n;
    payload = p;
    packetSize = pktSize-40;
    protoOvhd = 40;
    conn = NULL;
    board = new AckScoreboard(400);
    flowNo = n->GetFlowNo();

    maxSamples = 20;
    seqNos = new int[maxSamples];
    rttSamples = new float[maxSamples];
    msgInitTimes = new float[maxSamples];
    startTime = finishTime = 0;
    isActive = 0;
    thinkTime = think;
    n->SetFlow(flowNo, this);
}

void FQSender::Start(float t)
{
    Event *e = new Event(t, this, START_TIMER, NULL, t);
    net->AddEvent(e);
}

void FQSender::CleanState()
{
    bytesSent = 0;
    numRexmits = 0;
    rto = 3000;		/* 3secs - a large value */
    minRtt = rto;
    rttEstimate = 0;	  
    rttVarEstimate = 0;
    senderDone = 0;
    spacing = 0;
    numSamples = 0;
    board->CleanScoreboard();
}

void FQSender::RepeatedStart(int num)
{
    int i;
    float t = 0;
    for (i=0; i<num; i++) {
	t += ThinkTimeDist(thinkTime);	
	/* t += thinkTime*(0.7 + 0.6*((random() % 100)/100.0)); */
	Event *e = new Event(t, this, START_TIMER, NULL, t);
	net->AddEvent(e);
    }
}

/* sends a message -- could be a retransmission */
void FQSender::SendMessage(int seq, int size, float curr)
{
    Message *m = MakeMessage(flowNo, 0, size, conn, seq, 0, curr);
    m->conn->agents[0]->HandleMessage(m, curr);

    Event *retEv = new Event(curr + RTO(), this, TIMEOUT, (void *)seq, curr);
    net->AddEvent(retEv);
    numPackets++;
}

void FQSender::RechargeTimeout(float curr)
{
    Event *retEv = new Event(curr + RTO(), this, TIMEOUT, 0, curr);
    net->AddEvent(retEv);
    nextTimeout = curr + RTO();
}

void FQSender::RecordedSendMessage(int seq, int size, float curr)
{
    if (size == 0 && seq != 0) return;
#ifdef DEBUG
    printf("%f: message %d:%d (size %d) sent\n", curr, flowNo, seq, size);
#endif
    SendMessage(seq, size + protoOvhd, curr);
    bytesSent += size;
    board->RecordMsg(seq, size, 0, curr);
    board->AdvanceSeqNo();
    numUniquePackets++;
}

void FQSender::RegisterFlow()
{
    for (int i=0; i<conn->numAgents-1; i++) {
	LinkAgent *l = (LinkAgent *)conn->agents[i];
	l->RegisterFlow(flowNo);
    }
    ((GenericReceiver *)conn->agents[conn->numAgents-1])->RegisterFlow(flowNo);
}

void FQSender::UnregisterFlow()
{
    for (int i=0; i<conn->numAgents-1; i++) {
	LinkAgent *l = (LinkAgent *)conn->agents[i];
	l->UnregisterFlow(flowNo);
    }
    ((GenericReceiver *)conn->agents[conn->numAgents-1])->UnregisterFlow(flowNo);
}

void FQSender::HandleEvent(Event *e)
{
    int type = e->evType;
    float curr = e->time;

    if (type == START_TIMER) {
	/* interrupting a flow */
	if (isActive) {
	    FinishFlow(curr);
	}
	CleanState();
	startTime = curr;
	isActive = 1;	
	RegisterFlow();
	/* RecordedSendMessage(0, SendPacketSize(), curr); */
	RecordedSendMessage(0, 0, curr);
	/*	RecordedSendMessage(1, SendPacketSize(), curr + 0.001); */
	return;
    }

    if (senderDone) return;

    if (e->initTime < startTime) return;

    if (type == SEND_TIMER) {
	if (curr != nextXmit)
	    return;

	/* if (board->UnackedBytes() < myShare*rttEstimate) { */
	if (board->CanSendMsg()) {
	    int seq = board->NextSeqNo();
	    RecordedSendMessage(seq, SendPacketSize(), curr);
	}
	lastXmit = curr;
	nextXmit = curr + spacing;
	Event *nextEv = new Event(nextXmit, this, SEND_TIMER, NULL, curr);
	net->AddEvent(nextEv);
    }
    
    if (type == ENQUEUE_TIMER) {

	HandleMessage((Message *)(e->data), curr);

    } else if (type == TIMEOUT) {
	int pktSeqNo = (int)e->data;
	if (!board->Receivedp(pktSeqNo)) {
	    float oldrto = curr - e->initTime;
	    float newrto = RTO();
	    /* float newrto = oldrto; */
	    
	    if (newrto > 1.1*oldrto) {
		Event *retEv = new Event(e->initTime + RTO(), this, TIMEOUT,
					 (void *)pktSeqNo, e->initTime);
		net->AddEvent(retEv);
	    } else {
		/*		printf("timeout********** for %d:%d at %f\n", flowNo, pktSeqNo, curr); */
		SendMessage(pktSeqNo, board->MsgSize(pktSeqNo), curr);
		board->RecordResend(pktSeqNo, curr);
		numRexmits += board->MsgSize(pktSeqNo);
	    }
	} else if (type == RTT_TIMER) {
#if 0
	    if (qLen > 1.0) 
		UpdateSpacing(2*(packetSize + protoOvhd)/myShare, curr);
	    else
		UpdateSpacing((packetSize + protoOvhd)/myShare, curr);
#endif
	    
	    float bwRate;
	    if (qLen < 1)
		bwRate = myShare;
	    else if (qLen < 1.5)
		bwRate = myShare * (2 - qLen);
	    else
		bwRate = myShare/2;

	    UpdateSpacing((packetSize + protoOvhd)/bwRate, curr);
	    Event *nextEv = new Event(curr + minRtt, this, RTT_TIMER, NULL, curr);
	    net->AddEvent(nextEv);
	}
	
#if 0
	if (nextTimeout > curr)
	    return;
	assert(board->UnackedSegments() > 0);
	int pktSeqNo = board->OldestSeqNo();

#if 0
	/*	printf("timeout********** for %d:%d at %f\n", flowNo, pktSeqNo, curr); */
	SendMessage(pktSeqNo, board->MsgSize(pktSeqNo), curr);
	/*	rto = 2*rto; */
	RechargeTimeout(curr);
#endif
	
	    
	if (newrto > 1.2*oldrto) {
	    RechargeTimeout(newrto - oldrto);
	} else {
	    printf("timeout********** for %d:%d at %f\n", flowNo, pktSeqNo, curr);
	    SendMessage(pktSeqNo, board->MsgSize(pktSeqNo), curr);
	    RechargeTimeout(curr);	    
	    board->RecordResend(pktSeqNo);
	}
#endif
    }
}

void FQSender::UpdateSpacing(float newspacing, float currTime)
{
    if (spacing == 0) {
	if (newspacing > 0)
	    spacing = newspacing;
	nextXmit = currTime;
	Event *nextEv = new Event(nextXmit, this, SEND_TIMER, NULL, currTime);
	net->AddEvent(nextEv);
    } else if (newspacing > 0) {
	float oldspacing = spacing;
	/*	spacing = 0.9*spacing + 0.1*newspacing; */
	spacing = newspacing;
	float elapsed = currTime - lastXmit;
	float interval = nextXmit - lastXmit;
	float fraction = elapsed/interval;
	float oldNextXmit = nextXmit;
	if (fabs(oldNextXmit - nextXmit) > 0.01*spacing) {
	    printf("Flow %d changed spacing %f --> %f\n", flowNo, spacing, oldspacing);
	    nextXmit = currTime + (1-fraction)*2*spacing;
	    lastXmit = currTime - fraction*2*spacing;
	    if (nextXmit < currTime) {
		nextXmit = currTime;
	    }
	    Event *nextEv = new Event(nextXmit, this, SEND_TIMER, NULL, currTime);
	    net->AddEvent(nextEv);
	}
    }
}

void FQSender::RecordSample(int seq, float rtt, float currTime, float bwshare, float qlen)
{
    int i, j;
    int otherseq = ((seq % 2) ? (seq - 1) : (seq + 1));

    /*    myShare = 0.9*myShare + 0.1*bwshare; */
    myShare = 1.1*bwshare;
    qLen = qlen;

#if 0
    for (i=0; i<numSamples; i++) {
	if (seqNos[i] == otherseq) {
#if 0
	    float newspacing = rtt - rttSamples[i];
	    UpdateSpacing(newspacing, currTime);
#endif
	    if (fabs(msgInitTimes[i]-(currTime-rtt)) < 0.1) {
		float newspacing = rtt - rttSamples[i];
		UpdateSpacing(newspacing, currTime);
	    }
	    for (j=i; j<numSamples-1; j++) {
		seqNos[j] = seqNos[j+1];
		rttSamples[j] = rttSamples[j+1];
		msgInitTimes[j] = msgInitTimes[j+1];
	    }
	    numSamples--;
	    return;
	}
    }
    if (numSamples < maxSamples) {
	seqNos[numSamples] = seq;
	rttSamples[numSamples] = rtt;
	msgInitTimes[numSamples] = currTime - rtt;
	numSamples++;
    } else {
	for (i=0; i<numSamples-1; i++) {
	    seqNos[i] = seqNos[i+1];
	    rttSamples[i] = rttSamples[i+1];
	    msgInitTimes[i] = msgInitTimes[i+1];	    
	}
	seqNos[numSamples-1] = seq;
	rttSamples[numSamples-1] = rtt;
	msgInitTimes[numSamples-1] = currTime - rtt;
    }
#endif
}

void FQSender::FinishFlow(float curr)
{
    UnregisterFlow();
    senderDone = 1;
    if (!isActive) {
	finishTime = startTime + (curr - startTime)*(payload/bytesSent);	
    } else {
	isActive = 0;
	finishTime = curr;
    }
    TrackResponseTime(finishTime - startTime, bytesSent, numRexmits, numPackets, numUniquePackets);
}

void FQSender::HandleMessage(Message *m, float curr)
{
    if (board->Receivedp(m->seqno))
	return;
    
    board->RecordAck(m->seqno);
    /* RechargeTimeout(curr); */
#ifdef DEBUG
    printf("%f: message %d:%d ack received, rtt: %f\n", curr, m->flowNo, m->seqno, curr - m->startTime);
#endif
    if (board->UnackedBytes() == 0 && payload == bytesSent) {
	FinishFlow(curr);
	return;
    }
    
    if (m->seqno == 0) {
	
	rttEstimate = curr - m->startTime;
	/* initialize rttVarEstimate to something high */
	rttVarEstimate = (curr - m->startTime)/2;
	rto = rttEstimate + 4*rttVarEstimate;
	if (minRtt > rttEstimate)
	    minRtt = rttEstimate;

	UpdateSpacing((packetSize + protoOvhd)/m->bwshare, curr);

	Event *nextEv = new Event(curr + rttEstimate, this, RTT_TIMER, NULL, curr);
	net->AddEvent(nextEv);
	RecordSample(m->seqno, curr - m->startTime, curr, m->bwshare, m->qlen);
	
    } else {
	
	float newRtt = curr - m->startTime;
	float newRttVar = fabs(newRtt - rttEstimate);

	if (minRtt > newRtt)
	    minRtt = newRtt;
	rttEstimate = 0.875*rttEstimate + 0.125*newRtt;
	rttVarEstimate = 0.75*rttVarEstimate + 0.25*newRttVar;
	rto = rttEstimate + 4*rttVarEstimate;	
	RecordSample(m->seqno, curr - m->startTime, curr, m->bwshare, m->qlen);

#if 0
	    float bwRate;
	    if (qLen < 1)
		bwRate = myShare;
	    else if (qLen < 1.5)
		bwRate = myShare * (2.0 - qLen);
	    else
		bwRate = myShare/2;

	    UpdateSpacing((packetSize + protoOvhd)/bwRate, curr);
#endif
    }
    
    /* UpdateSpacing((packetSize + protoOvhd)/myShare, curr); */
    
#if 0
    int reordSeq = board->CheckFastRetransmit();
    if (reordSeq != -1) {
	/* perform a retransmission */
	SendMessage(reordSeq, board->MsgSize(reordSeq), curr);
	board->RecordResend(reordSeq);
	return;
    }
#endif
    delete m;
}
