#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include "tscp.h"

void FQFlow::InitFlow(int msgs, int segSize)
{
    flowId = -1;
    begin = end = numMessages = numBytes = 0;
    maxMessages = msgs;
    buffer = new (Message *)[msgs];
    startRound = finishRound = 0;
    segmentSize = segSize;
    inActiveHeap = 0;
    allocShare = 0;
}

void FQFlow::PrintStatus()
{
    printf("Flow: %d Num: %d Start: %f End: %f MIDS: ", flowId, numMessages, startRound, finishRound);
    for (int i=0; i<numMessages; i++)
	printf("%d ", buffer[(begin + i) % maxMessages]->seqno);
    printf("\n");
}

void FQFlow::StartFlow(int fId)
{
    flowId = fId;
    begin = end = numMessages = numBytes = 0;
}

int FQFlow::IsDormant()
{
    return (flowId == -1);
}

void FQFlow::EndFlow(Heap *szHeap, Heap *priorHeap, Heap *activeHeap, FQFlow **xmitFlow)
{
    flowId = -1;
    szHeap->RemoveItem(sizeHeapPos, this);
    if ((numMessages > 0) && (this != *xmitFlow))
	priorHeap->RemoveItem(priorHeapPos, this);
    else if (this == *xmitFlow)
	*xmitFlow = NULL;
    if (inActiveHeap) {
	activeHeap->RemoveItem(activeHeapPos, this);
	inActiveHeap = 0;
    }
}

int FQFlow::AddTail(Message *m, Heap *sizeHeap)
{
    if (numMessages == maxMessages)
	return 0;

    buffer[end] = m;
    end = (end + 1) % maxMessages;
    numMessages++;
    numBytes += m->size;
    ResizeFlow(sizeHeap);
    return 1;
}

Message *FQFlow::RemoveHead(Heap *priorHeap, Heap *sizeHeap)
{
    if (numMessages == 0)
	assert(0);

#ifdef DEBUG
    printf("Transmit: %f\n", finishRound);
#endif
    
    Message *m = buffer[begin];
    begin = (begin + 1) % maxMessages;
    numMessages--;
    numBytes -= m->size;
    ResizeFlow(sizeHeap);    
    if (numMessages == 0)
	begin = end = 0;
    else {
	startRound = finishRound;
	finishRound = startRound + buffer[begin]->size;
	priorHeap->InsertElem(finishRound, this, &priorHeapPos);
    }
    return m;
}

Message *FQFlow::PeepHead()
{
    return buffer[begin];
}

Message *FQFlow::RemoveTail()
{
    assert (numMessages > 0);
    Message *m = buffer[(end-1+maxMessages) % maxMessages];
    end = (end - 1 + maxMessages) % maxMessages;
    numMessages--;
    numBytes -= m->size;
    return m;
}

void FQFlow::ResizeFlow(Heap *h)
{
    h->ChangeCost(sizeHeapPos, maxMessages*segmentSize - numBytes);
}

void FQFlow::AddToSizeHeap(Heap *h)
{
    h->InsertElem(maxMessages*segmentSize - numBytes, this, &sizeHeapPos);
}

void FQFlow::AddToActiveHeap(Heap *h)
{
    if (!inActiveHeap) {
	inActiveHeap = 1;
	h->InsertElem(finishRound, this, &activeHeapPos);
    }
}

void FQFlow::SetStartRound(float num, Heap *h)
{
    startRound = num;
    finishRound = num + buffer[begin]->size;
    h->InsertElem(finishRound, this, &priorHeapPos);
}
