#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include "heap.h"

Heap::Heap()
{
    num_elems = 0;
    max_elems = 1000;
    nodes = new HeapNode[1000];
}

Heap::Heap(int size)
{
    num_elems = 0;
    max_elems = size;
    nodes = new HeapNode[size];
}

Heap::~Heap()
{
    delete nodes;
}

void Heap::UpdatePosPtr(int pos)
{
    if (nodes[pos].pos_ptr != NULL)
	*(nodes[pos].pos_ptr) = pos;
}

void Heap::PushUp(int pos)
{
    while (pos > 0) {
	int par = (pos-1)/2;
	if (nodes[pos].cost < nodes[par].cost) {
	    HeapNode t = nodes[par];
	    nodes[par] = nodes[pos];
	    nodes[pos] = t;
	    UpdatePosPtr(pos);
	    UpdatePosPtr(par);
	    pos = par;
	} else
	    return;
    }
}

void Heap::PushDown(int pos)
{
    int c1 = 2*pos + 1;
    int c2 = 2*pos + 2;

    while (c1 < num_elems) {

	if (c2 < num_elems) {

	    if ((nodes[pos].cost < nodes[c1].cost) &&
		(nodes[pos].cost < nodes[c2].cost))
		return;

	    if (nodes[c1].cost < nodes[c2].cost) {
		HeapNode t = nodes[c1];
		nodes[c1] = nodes[pos];
		nodes[pos] = t;
		UpdatePosPtr(pos);
		UpdatePosPtr(c1);
		pos = c1;
	    } else {
		HeapNode t = nodes[c2];
		nodes[c2] = nodes[pos];
		nodes[pos] = t;
		UpdatePosPtr(pos);
		UpdatePosPtr(c2);
		pos = c2;
	    }

	} else {

	    if (nodes[c1].cost < nodes[pos].cost) {
		HeapNode t = nodes[c1];
		nodes[c1] = nodes[pos];
		nodes[pos] = t;
		UpdatePosPtr(pos);
		UpdatePosPtr(c1);
		pos = c1;
	    } else
		return;
	    
	}

	c1 = 2*pos + 1;
	c2 = 2*pos + 2;

    }
    
}

void Heap::InsertElem(float c, void *it, int *pos_ptr)
{
    int pos = num_elems;
    assert(pos < max_elems);
    nodes[pos].cost = c;
    nodes[pos].item = it;
    nodes[pos].pos_ptr = pos_ptr;
    UpdatePosPtr(pos);
    num_elems++;
    PushUp(pos);
}

void *Heap::DeleteMin()
{
    void *ret;

    if (num_elems == 0)
	return NULL;

    ret = nodes[0].item;
    num_elems--;
    nodes[0] = nodes[num_elems];
    UpdatePosPtr(0);
    PushDown(0);
    
    return ret;
}

int Heap::IsEmpty()
{
    return (num_elems == 0);
}

float Heap::PeekMinCost()
{
    assert(num_elems > 0);
    return nodes[0].cost;
}

void Heap::ChangeCost(int pos, float new_cost)
{
    float old_cost = nodes[pos].cost;
    nodes[pos].cost = new_cost;

    if (old_cost > new_cost) {
	PushUp(pos);
    } else if (old_cost < new_cost) {
	PushDown(pos);
    }
}

void Heap::RemoveItem(int pos, void *it)
{
    assert(nodes[pos].item == it);
    assert(num_elems > 0);
    num_elems--;
    nodes[pos] = nodes[num_elems];
    UpdatePosPtr(pos);
    PushDown(pos);
}

void Heap::PrintHeap()
{
    int i;

    for (i=0; i<num_elems; i++)
	printf("%3.1f ", nodes[i].cost);
    printf("***********************************************\n");
}

#define TEST_ELEMS 1000
void TestHeap()
{
    int i, num;
    int mask[TEST_ELEMS];
    
    Heap *h = new Heap(2000);
    for (i=0; i<TEST_ELEMS; i++)
	mask[i] = 0;

    for (num=0; num<TEST_ELEMS; ) {
	int x = random() % TEST_ELEMS;
	if (!mask[x]) {
	    int *elem = (int *)malloc(sizeof(int));
	    int pos;
	    *elem = x;
	    h->InsertElem(x, elem, &pos);
	    num++;
	    if (num % 300 == 0)
		h->PrintHeap();
	    mask[x] = 1;
	}
    }
    printf("***********************************************\n");
    for (i=0; i<TEST_ELEMS; i++) {
	int *it = (int *) h->DeleteMin();
	printf("%d ", *it);
    }
    printf("\n");
	
    delete h;
}
