#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "tscp.h"

float factor = 1.0;
extern void ReadWorkload(Network *, char *);

float total = 0;
float totalSamples = 0;

void Send(Network *n, int flowno, int size, Connection *c, float t)
{
    static int seqno = 0;
    Message *m = MakeMessage(flowno, 0, size, c, seqno++, 0, t);
    Event *e = new Event(t, c->agents[0], ENQUEUE_TIMER, m, 0);
    n->AddEvent(e);
}

#include <math.h>

int RandSize(int mean)
{
    float shape = 1.25;
    float mode = mean*(shape - 1)/shape;
    float randVal = (random() % 100000)/100000.0;
    int retVal = (int)(mode/pow((1 - randVal), 1/shape));
    return retVal;
}

int FQStressTest(int argc, char **argv)
{
    if (argc != 1) {
	fprintf(stderr, "usage: tscp\n");
	exit(1);
    }

    Network n;
    FQLink *fq = new FQLink(&n, 40, 1250, 100, 1, 20, 100, 0);
    NullAgent *na = new NullAgent;
    Agent *agents[2] = { fq, na };
    Connection *c = MakeConnection(2, agents);

    for (int i=0; i<20; i++)
	fq->RegisterFlow(i);
    
    /* Message: flowno type size conn seqno time */
    for (int j=0; j<1000; j++)
	Send(&n, random() % 20, 100, c, 0.5*j);
#if 0
    Send(&n, 0, 100, c, 0.0);
    Send(&n, 1, 100, c, 0.5);
    Send(&n, 2, 100, c, 0.8);
    Send(&n, 0, 100, c, 0.95);
    Send(&n, 1, 100, c, 1.35);
    Send(&n, 2, 100, c, 2.0);
#endif
    while (n.Step());
}

int FileWorkload(int argc, char **argv)
{
    Network n;
    if (argc != 2) {
	fprintf(stderr, "usage: tscp workloadfile\n");
	exit(1);
    }
    ReadWorkload(&n, argv[1]);
    while (n.Step());



#if 0
    /* old stuff */
    FifoLink l1(1000, 10000, 1), l2(1000, 10000, 1),
	l3(1000, 1000, 1), l4(1000, 1000, 1),
	l5(1000, 10000, 1), l6(1000, 10000, 1);

    int numBytes;
    sscanf(argv[1], "%d", &numBytes);
    TscpSender s(numBytes);
    TscpReceiver r;

    Agent *agents1[] = {&l1, &l3, &l5, &r};
    Connection *c1 = MakeConnection(4, agents1);
    s.SetConnection(c1);

    Agent *agents2[] = {&l6, &l4, &l2, &s};
    Connection *c2 = MakeConnection(4, agents2);
    r.SetConnection(c2);

    s.Start(&n, 100);
#endif

}

void ScheduleStartFlow(Network *n, char *type, int numFlows, int s, int d,
		       float t, int flowSize, int pktSize,
		       LinkAgent **links,
		       SenderAgent **sender, GenericReceiver **receiver)
{
    Agent *agents1[4]; 
    Agent *agents2[4];
    static TscpGlobalStats *tscpStats = NULL;
    static TcpGlobalStats *tcpStats = NULL;

    if ((strcmp(type, "pcp") == 0) || (strcmp(type, "pcp-fq") == 0)) {
	if (tscpStats == NULL) {
	    tscpStats = new TscpGlobalStats(n, 100000, 5000);
	    n->SetTscpStats(tscpStats);
	}
	*sender = new TscpSender(n, pktSize*flowSize, pktSize, -1, tscpStats);
	*receiver = new GenericReceiver(n);
    } else if ((strcmp(type, "tcp") == 0) || (strcmp(type, "tcp-fq") == 0)) {
	if (tcpStats == NULL) {
	    tcpStats = new TcpGlobalStats(n, 100000, 5000);
	    n->SetTcpStats(tcpStats);
	}
	*sender = new TcpSender(n, pktSize*flowSize, pktSize, -1, tcpStats);
	*receiver = new TcpReceiver(n);	
    } else if (strcmp(type, "fq") == 0) {
	*sender = new FQSender(n, pktSize*flowSize, pktSize, -1);
	*receiver = new GenericReceiver(n);
    }

    agents1[0] = links[4*s];
    agents1[1] = links[4*numFlows];
    agents1[2] = links[4*d+2];
    agents1[3] = *receiver;
	    
    agents2[0] = links[4*d+3];
    agents2[1] = links[4*numFlows+1];
    agents2[2] = links[4*s+1];
    agents2[3] = *sender;

    Connection *c1 = MakeConnection(4, agents1);
    Connection *c2 = MakeConnection(4, agents2);
    (*sender)->SetConnection(c1);
    (*receiver)->SetConnection(c2);
    (*sender)->Start(t);
}

int ClosedLoopDriver(int argc, char **argv)
{
    Network n;
    float bbw, avgrtt;
    int pktSize, numFlows, flowSize, numRepeats;
    float thinkTime, startMonitor, endMonitor;
    LinkAgent **links;
    SenderAgent **senders;
    GenericReceiver **receivers;
    int protoOvhd = 40;
    
    if (argc != 10) {
	fprintf(stderr, "usage: tscp scheme bottleneckbw avgrtt pktsize numflows flowsize thinktime starttime endtime\n");
	exit(1);
    }
    sscanf(argv[2], "%f", &bbw);
    sscanf(argv[3], "%f", &avgrtt);
    sscanf(argv[4], "%d", &pktSize);
    sscanf(argv[5], "%d", &numFlows);
    sscanf(argv[6], "%d", &flowSize);
    sscanf(argv[7], "%f", &thinkTime);
    sscanf(argv[8], "%f", &startMonitor);
    sscanf(argv[9], "%f", &endMonitor);
    factor = 1.0;

    links = new (LinkAgent *)[2*(2*numFlows + 1)];
    senders = new (SenderAgent *)[numFlows];
    receivers = new (GenericReceiver *)[numFlows];

    float xmitdelay = (pktSize/bbw) + 2*(pktSize/(10*bbw))
		+ (protoOvhd/bbw) + 2*(protoOvhd/(10*bbw));
    int queueSize = (int)(bbw*avgrtt/pktSize);
 
    for (int j=0; j<numFlows; j++) {
	/*	float rtt = avgrtt - 1.0 + ((float)(random() % 200))/100.0; */
	float rtt = avgrtt - 10.0 + (float)(random() % 20);
	float delay = (rtt - xmitdelay - 2.0)/4;
	links[4*j] = new FifoLink(&n, 2*queueSize*pktSize, 10*bbw, delay, 0);
	links[4*j+1] = new FifoLink(&n, 2*queueSize*pktSize, 10*bbw, delay, 0);
	links[4*j+2] = new FifoLink(&n, 2*queueSize*pktSize, 10*bbw, delay, 0);
	links[4*j+3] = new FifoLink(&n, 2*queueSize*pktSize, 10*bbw, delay, 0);
    }

    if (strcmp(argv[1], "pcp") == 0) {
	links[4*numFlows] = new FifoLink(&n, queueSize*pktSize, bbw, 1, 1);
	links[4*numFlows+1] = new FifoLink(&n, queueSize*pktSize, bbw, 1, 0);
    } else if (strcmp(argv[1], "tcp") == 0) {
	links[4*numFlows] = new FifoLink(&n, queueSize*pktSize, bbw, 1, 1);
	links[4*numFlows+1] = new FifoLink(&n, queueSize*pktSize, bbw, 1, 0);
    } else if ((strcmp(argv[1], "fq") == 0)
	       || (strcmp(argv[1], "pcp-fq") == 0)
	       || (strcmp(argv[1], "tcp-fq") == 0)) {
	links[4*numFlows] = new FQLink(&n, queueSize*pktSize, pktSize, bbw, 1,
				       4*numFlows, (int)(2*(endMonitor/thinkTime)*numFlows), 1);
	links[4*numFlows+1] = new FQLink(&n, queueSize*pktSize, pktSize, bbw, 1,
					 4*numFlows, (int)(2*(endMonitor/thinkTime)*numFlows), 0);
    }

    float t = 0;
    while (t < endMonitor) {
	SenderAgent *s;
	GenericReceiver *r;
	t += ExpTimeDist(thinkTime);
	int flowNo = random() % numFlows;
	int randSz = RandSize(flowSize);
	ScheduleStartFlow(&n, argv[1], numFlows, flowNo, flowNo, t,
			  randSz, pktSize, links, &s, &r);
	if (t > startMonitor) {
	    total += ((float)randSz)/flowSize;
	    totalSamples++;
	}
    }
#if 0
    if (strcmp(argv[1], "pcp") == 0) {
	Agent *agents1[4]; 
	Agent *agents2[4]; 
	TscpGlobalStats *tscpStats = new TscpGlobalStats(&n, 1200, 5000);
	
	links[4*numFlows] = new FifoLink(&n, queueSize, bbw, 1);
	links[4*numFlows+1] = new FifoLink(&n, queueSize, bbw, 1);
	for (int i=0; i<numFlows; i++) {
	    senders[i] = new TscpSender(&n, pktSize*flowSize, pktSize, thinkTime, tscpStats);
	    receivers[i] = new GenericReceiver(&n);
	    agents1[0] = links[4*i];
	    agents1[1] = links[4*numFlows];
	    agents1[2] = links[4*i+2];
	    agents1[3] = receivers[i];
	    
	    agents2[0] = links[4*i+3];
	    agents2[1] = links[4*numFlows+1];
	    agents2[2] = links[4*i+1];
	    agents2[3] = senders[i];

	    Connection *c1 = MakeConnection(4, agents1);
	    Connection *c2 = MakeConnection(4, agents2);
	    senders[i]->SetConnection(c1);
	    receivers[i]->SetConnection(c2);
	    senders[i]->RepeatedStart(3*endMonitor/thinkTime);
	}
    } else if (strcmp(argv[1], "tcp") == 0) {
	Agent *agents1[4]; 
	Agent *agents2[4]; 
	TcpGlobalStats *tcpStats = new TcpGlobalStats(&n, 1200, 5000);
	
	links[4*numFlows] = new FifoLink(&n, queueSize, bbw, 1);
	links[4*numFlows+1] = new FifoLink(&n, queueSize, bbw, 1);
	for (int i=0; i<numFlows; i++) {
	    senders[i] = new TcpSender(&n, pktSize*flowSize, pktSize, thinkTime, tcpStats);
	    receivers[i] = new GenericReceiver(&n);
	    agents1[0] = links[4*i];
	    agents1[1] = links[4*numFlows];
	    agents1[2] = links[4*i+2];
	    agents1[3] = receivers[i];
	    
	    agents2[0] = links[4*i+3];
	    agents2[1] = links[4*numFlows+1];
	    agents2[2] = links[4*i+1];
	    agents2[3] = senders[i];

	    Connection *c1 = MakeConnection(4, agents1);
	    Connection *c2 = MakeConnection(4, agents2);
	    senders[i]->SetConnection(c1);
	    receivers[i]->SetConnection(c2);
	    senders[i]->RepeatedStart(3*endMonitor/thinkTime);
	}
    } else if (strcmp(argv[1], "fq") == 0) {
	Agent *agents1[4]; 
	Agent *agents2[4]; 
	
	links[4*numFlows] = new FQLink(&n, queueSize, pktSize, bbw, 1, 2*numFlows, 2*numFlows);
	links[4*numFlows+1] = new FQLink(&n, queueSize, pktSize, bbw, 1, 2*numFlows, 2*numFlows);
	for (int i=0; i<numFlows; i++) {
	    senders[i] = new FQSender(&n, pktSize*flowSize, pktSize, thinkTime);
	    receivers[i] = new GenericReceiver(&n);
	    agents1[0] = links[4*i];
	    agents1[1] = links[4*numFlows];
	    agents1[2] = links[4*i+2];
	    agents1[3] = receivers[i];
	    
	    agents2[0] = links[4*i+3];
	    agents2[1] = links[4*numFlows+1];
	    agents2[2] = links[4*i+1];
	    agents2[3] = senders[i];

	    Connection *c1 = MakeConnection(4, agents1);
	    Connection *c2 = MakeConnection(4, agents2);
	    senders[i]->SetConnection(c1);
	    receivers[i]->SetConnection(c2);
	    senders[i]->RepeatedStart(3*endMonitor/thinkTime);
	}
	
    } else
	assert(0);
#endif
    
    Event *ev = new Event(startMonitor, NULL, START_STATS, NULL, 0);
    n.AddEvent(ev);
    ev = new Event(endMonitor, NULL, END_STATS, NULL, 0);
    n.AddEvent(ev);

    while (n.Step());
}



int FixedFlows(int argc, char **argv)
{
    Network n;
    float bbw, avgrtt;
    int pktSize, numFlows, flowSize, numRepeats;
    float thinkTime, startMonitor, endMonitor;
    LinkAgent **links;
    SenderAgent **senders;
    GenericReceiver **receivers;
    int protoOvhd = 40;
    
    if (argc != 9) {
	fprintf(stderr, "usage: tscp scheme bottleneckbw avgrtt pktsize numflows flowsize starttime endtime\n");
	exit(1);
    }
    sscanf(argv[2], "%f", &bbw);
    sscanf(argv[3], "%f", &avgrtt);
    sscanf(argv[4], "%d", &pktSize);
    sscanf(argv[5], "%d", &numFlows);
    sscanf(argv[6], "%d", &flowSize);
    sscanf(argv[7], "%f", &startMonitor);
    sscanf(argv[8], "%f", &endMonitor);

    links = new (LinkAgent *)[2*(2*numFlows + 1)];
    senders = new (SenderAgent *)[numFlows];
    receivers = new (GenericReceiver *)[numFlows];

    float xmitdelay = (pktSize/bbw) + 2*(pktSize/(10*bbw))
		+ (protoOvhd/bbw) + 2*(protoOvhd/(10*bbw));
    int queueSize = (int)(0.7*bbw*avgrtt/pktSize);
 
    for (int j=0; j<numFlows; j++) {
	/*	float rtt = avgrtt - 1.0 + ((float)(random() % 200))/100.0; */
	float rtt = avgrtt;
	float delay = (rtt - xmitdelay - 2.0)/4;
	links[4*j] = new FifoLink(&n, 2*queueSize*pktSize, 10*bbw, delay, 0);
	links[4*j+1] = new FifoLink(&n, 2*queueSize*pktSize, 10*bbw, delay, 0);
	links[4*j+2] = new FifoLink(&n, 2*queueSize*pktSize, 10*bbw, delay, 0);
	links[4*j+3] = new FifoLink(&n, 2*queueSize*pktSize, 10*bbw, delay, 0);
    }

    if (strcmp(argv[1], "pcp") == 0) {
	links[4*numFlows] = new FifoLink(&n, queueSize*pktSize, bbw, 1, 1);
	links[4*numFlows+1] = new FifoLink(&n, queueSize*pktSize, bbw, 1, 0);
    } else if (strcmp(argv[1], "tcp") == 0) {
	links[4*numFlows] = new FifoLink(&n, queueSize*pktSize, bbw, 1, 1);
	links[4*numFlows+1] = new FifoLink(&n, queueSize*pktSize, bbw, 1, 0);
    } else if (strcmp(argv[1], "fq") == 0) {
	links[4*numFlows] = new FQLink(&n, queueSize*pktSize, pktSize, bbw, 1,
				       4*numFlows, (int)(2*(endMonitor/thinkTime)*numFlows), 1);
	links[4*numFlows+1] = new FQLink(&n, queueSize*pktSize, pktSize, bbw, 1,
					 4*numFlows, (int)(2*(endMonitor/thinkTime)*numFlows), 0);
    }

    SenderAgent *s;
    GenericReceiver *r;
    for (int k=0; k<numFlows-1; k++) {
	ScheduleStartFlow(&n, argv[1], numFlows, k, k, 0,
			  flowSize, pktSize, links, &s, &r);
    }
    ScheduleStartFlow(&n, argv[1], numFlows, numFlows-1, numFlows-1, 0,
		      flowSize/2, pktSize, links, &s, &r);

    Event *ev = new Event(startMonitor, NULL, START_STATS, NULL, 0);
    n.AddEvent(ev);
    ev = new Event(endMonitor, NULL, END_STATS, NULL, 0);
    n.AddEvent(ev);

    while (n.Step());    
    
}

int main(int argc, char **argv)
{
    
    /* FQStressTest(argc, argv); */
    /* FileWorkload(argc, argv); */
    ClosedLoopDriver(argc, argv);
    /* FixedFlows(argc, argv); */
}
