#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include "tscp.h"

Network::Network()
{
    currTime = 0;
    events = new Heap(100000);
    numFlows = numLinks = 0;
    history = NULL;
    maxFlows = maxLinks = 10000;
    links = new (LinkAgent *)[maxLinks];
    senders = new (SenderAgent *)[maxFlows];
    tscpStats = NULL;
    tcpStats = NULL;
}

void Network::AddEvent(Event *e)
{
    events->InsertElem(e->time, e, NULL);
}

void Network::HandleEvent(Event *e)
{
    int i;
    int type = e->evType;
    static float start;

    if (type == START_STATS) {
	EnableStats();
	for (i=0; i<numLinks; i++)
	    links[i]->EnableStats(e->time);
	start = e->time;
	if (tscpStats)
	    tscpStats->CleanStats();
    } else if (type == END_STATS) {
	if (tscpStats)
	    tscpStats->StopAllFlows(e->time);
	if (tcpStats)
	    tcpStats->StopAllFlows(e->time);
	PrintStats(e->time - start);
#if 0
	for (i=0; i<numLinks; i++)
	    links[i]->PrintStats(e->time);
#endif
	links[numLinks-2]->PrintStats(e->time);
	if (tscpStats)
	    tscpStats->PrintStats();
	exit(1);
    } else
	assert(0);
}

int Network::Step()
{
    static int nextEv = 1000;
    Event *e = (Event *)events->DeleteMin();
    if (e == NULL)
	return 0;
    if (e->agent == NULL)
	HandleEvent(e);
    else
	e->agent->HandleEvent(e);
    currTime = e->time;
    if (currTime > nextEv) {
	/* fprintf(stderr, "."); */
	nextEv += 1000;
    }
    delete e;
    return 1;
}

int Network::GetFlowNo()
{
    assert (numFlows < maxFlows);
    return numFlows++;
}

int Network::GetLinkNo()
{
    assert (numLinks < maxLinks);
    return numLinks++;
}

void Network::RecordHistory(Agent *s, Agent *d, float bw, float minrtt)
{
    HistoryRecord **prev = &history;
    HistoryRecord *elem = history;
    while (elem != NULL) {
	if (elem->s == s && elem->d == d) {
	    elem->bw = 0.9*elem->bw + 0.1*bw;
	    elem->minrtt = (elem->minrtt > minrtt ? minrtt : elem->minrtt);
#ifdef DEBUG
	    printf("history: %d-->%d is %f\n",
		   ((LinkAgent *)s)->LinkId(), ((LinkAgent *)d)->LinkId(), elem->bw);
#endif
	    return;
	}
	prev = &elem->next;
	elem = elem->next;
    }
    elem = *prev = new HistoryRecord;
    elem->s = s;
    elem->d = d;
    elem->bw = bw;
    elem->minrtt = minrtt;
    elem->next = NULL;
}

float Network::LookupHistory(Agent *s, Agent *d, float *minrtt)
{
    HistoryRecord *elem = history;
    while (elem != NULL) {
	if (elem->s == s && elem->d == d) {
	    *minrtt = elem->minrtt;
	    return elem->bw;
	}
	elem = elem->next;
    }
    return 0.0;
}
