#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include "stats.h"

StatsObject::StatsObject(int num)
{
    maxVals = num;
    numVals = 0;
    vals = new float [num];
    sorted = 0;
}

void StatsObject::RecordObservation(float v)
{
    sorted = 0;
    if (numVals == maxVals) {
	maxVals = 2*maxVals;
	float *newvals = new float[maxVals];
	for (int i=0; i<numVals; i++)
	    newvals[i] = vals[i];
	delete [] vals;
	vals = newvals;
    }
    vals[numVals++] = v;
}

static int floatcompare(const void *e1, const void *e2)
{
    float f1 = *(float *)e1;
    float f2 = *(float *)e2;
    if (f1 < f2)
	return -1;
    else if (f1 == f2)
	return 0;
    else
	return 1;
}

float StatsObject::QueryPercentile(int perc)
{
    if (!sorted) {
	qsort(vals, numVals, sizeof(float), floatcompare);
	sorted = 1;
    }
    return vals[(numVals*perc)/100];
}

void StatsObject::Print()
{
    for (int i=0; i<numVals; i++)
	printf("%f\n", vals[i]);
}

#if 0
int main(int argc, char **argv)
{
    StatsObject stats(100);
    for (int i=0; i<400; i++) {
	int v = random() % 500;
	printf("%f\n", (float)v);
	stats.RecordObservation((float)v);
    }
    printf("****\n");
    stats.Print();
    printf("%d : %f\n", 20, stats.QueryPercentile(20));
    printf("****\n");
    stats.Print();
}
#endif
