#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include "tscp.h"

/* #define DEBUG */
extern float factor;

#ifdef TRACK_BWS
static FILE *ratesFile = NULL;
#endif

TscpSender::TscpSender(Network *n, int p, int pktSize, float think, TscpGlobalStats *stats)
{
#ifdef TRACK_BWS    
    if (ratesFile == NULL)
	ratesFile = fopen("bwrates", "w");
#endif
    net = n;
    payload = p;
    bytesSent = 0;
    packetSize = pktSize-40;
    protoOvhd = 40;
    conn = NULL;

    numDrops = 0;
    flowNo = n->GetFlowNo();

    startTime = finishTime = 0;
    thinkTime = think;
    n->SetFlow(flowNo, this);
    globStats = stats;
    msgSizes = new Samples(4000, 100);
    uRates = new Samples(4000, 100);
    qSizes = new Samples(4000, 100);
}

void TscpSender::RepeatedStart(int num)
{
    int i;
    float t = 0;
    for (i=0; i<num; i++) {
	t += ThinkTimeDist(thinkTime);
	/*	t += thinkTime*(0.7 + 0.6*((random() % 100)/100.0)); */
	Event *e = new Event(t, this, START_TIMER, NULL, t);
	net->AddEvent(e);
    }
}

void TscpSender::CleanState()
{
    rttEstimate = 3000;	  /* 3secs - a large value */
    rttVarEstimate = 0;
    uniformRate = 0;
    currBurstRate = 0;
    lastXmitTime = nextXmitTime = 0;
    senderDone = 0;
    bytesSent = 0;
    board = new AckScoreboard(800);
    dropSeq = new int[200];
    board->CleanScoreboard();
    numDrops = 0;
    burstMsgsSent = 0;
    burstMsgsRcvd = 0;
    burstPhase = 0;
    msgSizes->Clean();
    uRates->Clean();
    qSizes->Clean();
    compTime = -3000;
    numRexmits = 0;
}

void TscpSender::Start(float t)
{
    Event *e = new Event(t, this, START_TIMER, NULL, t);
    net->AddEvent(e);
}

/* sends a message -- could be a retransmission */
void TscpSender::SendMessage(int seq, int size, float curr, int burst)
{
    Message *m = MakeMessage(flowNo, 0, size, conn, seq, 0, curr);
    if (burst)
	m->burstMsg = 1;
    m->conn->agents[0]->HandleMessage(m, curr);

    msgSizes->RecordSample(curr, size);
#if 1
    if ((board->NextSeqNo() != seq) && (board->TimeSent(seq) > curr - minRtt))
	assert(0);
#endif
    board->RecordSendTime(seq, curr);

    Event *retEv = new Event(curr + RTO(), this, TIMEOUT, (void *)seq, curr);
    net->AddEvent(retEv);
    numPackets++;
#ifdef DEBUG
    printf("%f: message %d:%d (size %d) sent\n", curr, flowNo, seq, size);
#endif
    /*    printf("message %d sent at %f\n", seq, curr); */
}

/* sends a message for the first time, remember state
   regarding the message in the scoreboard. */
void TscpSender::RecordedSendMessage(int seq, int size, float curr, int burst)
{
    if (size == 0) return;
#if 0
    if ((size == 0) && (board->UnackedBytes() == 0) && (payload == bytesSent))
	return;
#endif
    SendMessage(seq, size + protoOvhd, curr, burst);
    bytesSent += size;
    board->RecordMsg(seq, size, 0, curr);
    board->AdvanceSeqNo();
    numUniquePackets++;    
}

void TscpSender::RegisterFlow()
{
    for (int i=0; i<conn->numAgents-1; i++) {
	LinkAgent *l = (LinkAgent *)conn->agents[i];
	l->RegisterFlow(flowNo);
    }
    ((GenericReceiver *)conn->agents[conn->numAgents-1])->RegisterFlow(flowNo);
}

void TscpSender::UnregisterFlow()
{
    for (int i=0; i<conn->numAgents-1; i++) {
	LinkAgent *l = (LinkAgent *)conn->agents[i];
	l->UnregisterFlow(flowNo);
    }
    ((GenericReceiver *)conn->agents[conn->numAgents-1])->UnregisterFlow(flowNo);
}

void TscpSender::RegisterDrop(int seqno)
{
    dropSeq[numDrops] = seqno;
    numDrops++;
}

int TscpSender::CheckDropQueue()
{
    int seqno;
    if (numDrops == 0)
	return -1;
    seqno = dropSeq[0];
    numDrops--;
    for (int i=0; i<numDrops; i++)
	dropSeq[i] = dropSeq[i+1];
    return seqno;
}

int TscpSender::InDropQueue(int seq)
{
    for (int i=0; i<numDrops; i++)
	if (dropSeq[i] == seq)
	    return 1;
    return 0;
}

void TscpSender::RestartBurst(int pktSeqNo, float curr)
{
    for (int i=0; i<burstMsgsSent; i++) 
	if (burstMsgSeqno[i] == pktSeqNo) {
	    burstPhase--;
	    globStats->RecordBurstSuccess((burstMsgSize + protoOvhd)/burstSpacing, 0,
					  0, burstStartTime + 2*burstSpacing + rttEstimate/4);
	    StartNewBurst(curr);
	    globStats->TrackBPLoss();
	    break;
	}
}
			      

void TscpSender::HandleEvent(Event *e)
{
    int type = e->evType;
    float curr = e->time;

    if (type == BURST_STATS_ON) {
	float *p = (float *)&(e->data);
	globStats->StartBurstStats(curr, *p);
	currBurstRate = *p;
#ifdef TRACK_BWS	
	fprintf(ratesFile, "%d %f %f\n", flowNo, curr, uniformRate + currBurstRate);
	fflush(ratesFile);
#endif
    } else if (type == BURST_STATS_OFF) {
	float *p = (float *)&(e->data);
	globStats->StopBurstStats(curr, *p);
	currBurstRate = 0;
#ifdef TRACK_BWS	
	fprintf(ratesFile, "%d %f %f\n", flowNo, curr, uniformRate + currBurstRate);
	fflush(ratesFile);
#endif
    }
	
    if (type == START_TIMER) {
	if (isActive)
	    FinishFlow(curr);
	CleanState();
	startTime = curr;
	isActive = 1;
	RegisterFlow();

	Agent *s = conn->agents[0];
	Agent *d = conn->agents[conn->numAgents - 2];
	float minrttEst;
	float bwest = net->LookupHistory(s, d, &minrttEst);
	if (bwest > 0) {
	    rttEstimate = minrttEst;
	    rttVarEstimate = rttEstimate/2;
	}
	
	RecordedSendMessage(0, SendPacketSize(), curr, 0);
	uniformRate = 0.001;
	uRates->RecordSample(curr, uniformRate);
	lastXmitTime = curr;
	nextXmitTime = curr + SendInterval();
	globStats->StartFlowStats(this);
	return;
    }
    
    if (senderDone) return;

    if (e->initTime < startTime) return;
    
    if (type == ENQUEUE_TIMER) {

	HandleMessage((Message *)(e->data), curr);

    } else if (type == SEND_TIMER) {
	/* event has been overruled by updates to the underlying rate */
	if (curr != nextXmitTime)
	    return;

	if (board->CanSendMsg()) {
	    int seq = CheckDropQueue();
	    if (seq != -1 && !board->Receivedp(seq)) {
		SendMessage(seq, board->MsgSize(seq), curr, 0);
		numRexmits += board->MsgSize(seq);
	    }
	    else {
		if (bytesSent < payload) {
		    seq = board->NextSeqNo();
		    RecordedSendMessage(seq, SendPacketSize(), curr, 0);
		}
	    }
	}
	lastXmitTime = curr;
	nextXmitTime = curr + SendInterval();

	if (uniformRate > 0) {
	    Event *nextEv = new Event(curr + SendInterval(), this, SEND_TIMER, NULL, curr);
	    net->AddEvent(nextEv);
	}
    } else if (type == BURST_START_TIMER) {
	
	StartNewBurst(curr);
	
    } else if (type == BURST_TIMER) {

	SendBurstMessage(curr);
	
    } else if (type == TIMEOUT) {
	int pktSeqNo = (int)e->data;
	if (!board->Receivedp(pktSeqNo)) {
	    if (curr - e->initTime > 0.9*RTO()) {
#ifdef DEBUG
		printf("timeout********** for %d:%d at %f\n", flowNo, pktSeqNo, curr);
#endif
		if ((board->TimeSent(pktSeqNo) <= e->initTime) && !InDropQueue(pktSeqNo)) {
		    if (pktSeqNo != 0 && bytesSent < payload)
			RegisterDrop(pktSeqNo);
		    else
			SendMessage(pktSeqNo, board->MsgSize(pktSeqNo), curr, 0);
		    /* check whether we have to cancel a burst & start a new one */
		    RestartBurst(pktSeqNo, curr);
		}
	    } else {
		Event *retEv = new Event(e->initTime + RTO(), this, TIMEOUT,
					 e->data, e->initTime);
		net->AddEvent(retEv);
	    }
	}
    } else if (type == RTT_TIMER) {
#if 0
	float qmax = maxRtt - minRtt;
	float newRate = uniformRate - uniformRate*(qmax/minRtt)*factor;

	if (newRate <= 0) newRate = 0.001; /* some small rate */
#ifdef DEBUG
	printf("%f: flow %d underlying rate %f --> %f\n", curr, flowNo, uniformRate, newRate);
#endif
	UpdateUnderlyingRate(newRate, curr);
	maxRtt = minRtt;
	Event *nextEv = new Event(curr + minRtt*factor, this, RTT_TIMER, NULL, curr);
	net->AddEvent(nextEv);
#endif
    } 
}

/* not used... */
static float powint(int base, int exp)
{
    int i;
    float ret = 1;
    if (exp > 0) {
	for (i=0; i<exp; i++)
	    ret = base*ret;
	return ret;
    } else if (exp < 0) {
	for (i=0; i<exp; i++)
	    ret = ret/base;
    }
    return ret;
}

/* sends one message for the burst process */
void TscpSender::SendBurstMessage(float curr)
{
    if (board->CanSendMsg()) {
	int seq = board->NextSeqNo();
	float burstRate;

	if (burstMsgsSent == 0) {
	    
	    if (burstPhase >= 0)
		burstRate = pow(2.0, burstPhase)*((packetSize + protoOvhd)/minRtt);
	    else
		burstRate = (packetSize + protoOvhd)/minRtt;
	
	    burstMsgSize = (int)((burstRate*(minRtt/2.0))/5.0 - protoOvhd + 0.5);

	    if (burstMsgSize > packetSize) {
		burstMsgSize = packetSize;
		burstSpacing = (packetSize + protoOvhd)/burstRate;
	    } else {
		burstSpacing = minRtt/10.0;
	    }

#ifdef DEBUG
	    printf("%f: flow %d burst start for test rate %f\n", curr, flowNo, burstRate);
#endif
	    void **p = (void **)&burstRate;
	    Event *nextEv = new Event(curr + rttEstimate/4, this,
				      BURST_STATS_ON, *p, curr);
	    net->AddEvent(nextEv);
	    nextEv = new Event(curr + rttEstimate/4 + 5*burstSpacing, this,
			       BURST_STATS_OFF, *p, curr);
	    net->AddEvent(nextEv);

	    burstStartTime = curr;
#if 0
	    /* setup a smaller timeout for cancelling the burst */
	    Event *nextEv = new Event(curr + 3*minRtt, this, BURST_TIMEOUT, seq, curr);
	    net->AddEvent(nextEv);
#endif
	}
	int msgSize = (((payload - bytesSent) > burstMsgSize)
		       ? burstMsgSize : (payload - bytesSent));
	
	RecordedSendMessage(seq, msgSize, curr, 1);
	burstMsgSeqno[burstMsgsSent] = seq;
	burstMsgsSent++;

	if (burstMsgsSent < 5) {
	    Event *nextEv = new Event(curr + burstSpacing, this, BURST_TIMER, NULL, curr);
	    net->AddEvent(nextEv);
	}	
    } else {
	
	Event *nextEv = new Event(curr + minRtt/10, this, BURST_TIMER, NULL, curr);
	net->AddEvent(nextEv);
	printf("couldn't send message************************\n");
	
    }
}

/* updates underlying rate, readjusts the next timer, and sets up
   a virtual lastXmitTime to reflect the new rate */
void TscpSender::UpdateUnderlyingRate(float newRate, float curr)
{
    float oldRate = uniformRate;
    float elapsed = curr - lastXmitTime;
    float interval = nextXmitTime - lastXmitTime;

#ifdef TRACK_BWS
    fprintf(ratesFile, "%d %f %f\n", flowNo, curr, newRate + currBurstRate);
    fflush(ratesFile);
#endif
    uRates->RecordSample(curr, newRate);
    globStats->UpdateURateStats(oldRate, newRate, curr);
    uniformRate = newRate;
    if (uniformRate > 0.1) {
	nextXmitTime = curr + (1 - elapsed/interval)*(packetSize/newRate);
	// adjust the lastXmitTime so that multiple updates can take place
	lastXmitTime = curr - (elapsed/interval)*(packetSize/newRate);
	Event *nextEv = new Event(nextXmitTime, this, SEND_TIMER, NULL, curr);
	net->AddEvent(nextEv);
    }
}

static void LSE(float *y, float *x, int num, float *beta, float *offset, float *stderror)
{
    float my = 0, mx = 0;
    int i, j;

    for (i=0; i<num; i++) {
	mx += x[i];
	my += y[i];
    }
    mx = mx/num;
    my = my/num;

    float numer = 0, denom = 0;
    for (i=0; i<num; i++) {
	numer += (x[i] - mx)*(y[i] - my);
	denom += (x[i] - mx)*(x[i] - mx);
    }
    *beta = numer/denom;
    *offset = my - (*beta)*mx;

    float sigma = 0.0;
    for (i=0; i<num; i++) {
	float ecap = (y[i] - my) - (*beta)*(x[i] - mx);
	sigma += ecap*ecap;
    }
    sigma = sigma/(num - 1);
    denom = 0;
    for (i=0; i<num; i++)
	denom += (x[i] - mx)*(x[i] - mx);
    *stderror = sqrt(sigma/denom);
}

static float ranf()
{
    return ((float)(random() % 100000))/100000.0;
}

static float gauss_draw(float mean, float stddev)
{
    float x1, x2, w, y1, y2;
 
    do {
	x1 = 2.0 * ranf() - 1.0;
	x2 = 2.0 * ranf() - 1.0;
	w = x1 * x1 + x2 * x2;
    } while (w >= 1.0);

    w = sqrt((-2.0 * log(w))/w);
    y1 = x1 * w;
    y2 = x2 * w;
    return (y1*stddev) + mean;
}

static float uniform_draw(float mean, float stddev)
{
    float low = mean - sqrt(3)*stddev;
    float high = mean + sqrt(3)*stddev;
    return ranf()*(high - low) + low;
}

int TscpSender::CheckBurstFailure()
{
   float beta, offset, stderror;
   float xarray[] = {0.0, 1.0, 2.0, 3.0, 4.0};
   LSE(burstMsgRtt, xarray, 5, &beta, &offset, &stderror);
   float factor;

#if 1
   if (offset > 1.5*minRtt)
       factor = 0.01;
   else if (offset > 1.4*minRtt)
       factor = 0.02;
   else if (offset > 1.3*minRtt)
       factor = 0.03;
   else if (offset > 1.2*minRtt)
       factor = 0.04;
   else
       factor = 0.05;
#endif

   /*   factor = 0.03; */
   /* float qszInc = uniform_draw(beta, stderror); */
   float qszInc;
   if (burstPhase > 0)
       qszInc = beta + stderror;
   else
       qszInc = uniform_draw(beta, stderror);
   if (qszInc*(rttEstimate/burstSpacing) < factor*minRtt)
       return 0;
   else
       return 1;
}

#if 0
int TscpSender::CheckBurstFailure()
{

#if 0
   if ((offset + 13*(beta + stderror)) < 1.1*minRtt)
       return 0;
   else
       return 1;
   
   if (uniform_draw(beta, stderror) > rttEstimate/500)
       return 1;
   else
       return 0;

#endif
   double pct = 0.0;
   double pdtDenominator = 0.0;
   double pdtNumerator = 0.0;
   double average = 0.0;
   int count = 5;
   int i;

   for (i=1; i<count; i++) {
       pct += burstMsgRtt[i] > burstMsgRtt[i-1] ? 1.0 : 0.0;
       pdtDenominator += fabs (burstMsgRtt[i] - burstMsgRtt[i-1]);
       if (burstMsgRtt[i] > burstMsgRtt[i-1])
	   pdtNumerator += (burstMsgRtt[i] - burstMsgRtt[i-1]);
   }


   pct = pct / (count-1);
   
   double pdt;
   if (pdtDenominator > 0.0)
       pdt = (burstMsgRtt[count-1] - burstMsgRtt[0]) / pdtDenominator;
   else
       pdt = 0.0;
   return pct > 0.55 && pdt > 0.4;

#if 0
   pdt = pdtNumerator/pdtDenominator;

   for (i=0; i<count; i++)
       average += burstMsgRtt[i];

   average = average/count;
   
   if (pdtDenominator/average < 0.02)
       return 0;

   if (pct >= 0.75 && pdt > 0.5)
       return 1;

   if (pdt > 0.75)
       return 1;

   if (pdt < 0.25)
       return 0;

   return (((float)(random() % 1000))/2000.0 < (pdt - 0.5));
   /* return pct >= 0.4 && pdt > 0.5; */
   /*   return (((float)(random() % 1000))/1000.0 < pdt); */
   /*   return pdt > 0.6; */
#endif
}
#endif

void TscpSender::DebugCheckBurstFailure()
{
   double pct = 0.0;
   double pdtDenominator = 0.0;
   double pdtNumerator = 0.0;   
   int count = 5;

   for (int i=1; i<count; i++) {
       pct += burstMsgRtt[i] > burstMsgRtt[i-1] ? 1.0 : 0.0;
       pdtDenominator += fabs (burstMsgRtt[i] - burstMsgRtt[i-1]);
       if (burstMsgRtt[i] > burstMsgRtt[i-1])
	   pdtNumerator += (burstMsgRtt[i] - burstMsgRtt[i-1]);
   }

   pct = pct / (count-1);
   
   double pdt;
#if 0
   if (pdtDenominator > 0.0)
       pdt = (burstMsgRtt[count-1] - burstMsgRtt[0]) / pdtDenominator;
   else
       pdt = 0.0;
#endif
   pdt = pdtNumerator/pdtDenominator;

   float beta, ignore, stderror;
   float xarray[] = {0.0, 1.0, 2.0, 3.0, 4.0};
   LSE(burstMsgRtt, xarray, 5, &beta, &ignore, &stderror);

   /* return pct >= 0.4 && pdt > 0.5; */
   printf("(%f %f %f %f %f : %f) %f %f (%f:%f)\n", burstMsgRtt[0], burstMsgRtt[1],
	  burstMsgRtt[2], burstMsgRtt[3], burstMsgRtt[4], minRtt,
	  pct, pdt, beta, stderror);
}

void TscpSender::StartNewBurst(float curr)
{
    burstMsgsSent = burstMsgsRcvd = 0;
    if (burstPhase >= 0) {
#if 0
	float oldRate = uRates->LookupSample(burstStartTime);
	if (oldRate > 1.5*uniformRate) {
	    Event *nextEv = new Event(curr + minRtt, this, BURST_START_TIMER, NULL, curr);
	    net->AddEvent(nextEv);
	}
#endif
	SendBurstMessage(curr);
    } else {
	float waitTime = minRtt*(pow(2.0, -burstPhase) - 1.5);
	/* float waitTime = minRtt*(pow(2.0, 2) - 1.5); */
	Event *nextEv = new Event(curr + waitTime, this, BURST_TIMER, NULL, curr);
	net->AddEvent(nextEv);
    }
}

void TscpSender::ProcessBurstResults(float curr)
{
    float oldRate = uRates->LookupSample(burstStartTime);

#if 0
    if (burstMsgRtt[4] > 1.3*minRtt) {/* || oldRate > 1.2*uniformRate) { */
#ifdef DEBUG
	printf("%f: flow %d burst failure\n", curr, flowNo);
#endif
	globStats->RecordBurstSuccess((burstMsgSize + protoOvhd)/burstSpacing, 0,
				      0, burstStartTime + 2*burstSpacing + rttEstimate/4);
    } 
#endif
    if (CheckBurstFailure()) {
#ifdef DEBUG
	printf("%f: flow %d burst failure\n", curr, flowNo);
#endif
	burstPhase--;
	globStats->RecordBurstSuccess((burstMsgSize + protoOvhd)/burstSpacing, 0,
				      0, burstStartTime + 2*burstSpacing + rttEstimate/4);
    } else {
	float oldRate = uniformRate;
	float burstRate = (burstMsgSize + protoOvhd)/burstSpacing;
	globStats->RecordBurstSuccess((burstMsgSize + protoOvhd)/burstSpacing, 1,
				      0, burstStartTime + 2*burstSpacing + rttEstimate/4);
	UpdateUnderlyingRate(uniformRate + 0.9*burstRate, curr);
#ifdef DEBUG
	printf("%f: flow %d burst success at %f rate, uniform rate: %f --> %f\n",
	       curr, flowNo, burstRate, oldRate, uniformRate);
#endif
	if (burstPhase >= 0)
	    burstPhase++;
	else
	    burstPhase = 0;
    }
}

void TscpSender::FinishFlow(float curr)
{
    UnregisterFlow();
    senderDone = 1;
    if (!isActive) {
	net->RecordHistory(conn->agents[0], conn->agents[conn->numAgents - 2], uniformRate, minRtt);
	finishTime = startTime + (curr - startTime)*(payload/bytesSent);	
    } else {
	isActive = 0;
	finishTime = curr;
    }
    globStats->StopFlowStats(this);    
    TrackResponseTime(finishTime - startTime, bytesSent, numRexmits, numPackets, numUniquePackets);
    delete board;
    delete dropSeq;
}

void TscpSender::UCompensate(float curr)
{
    /*    float prevQmax = qSizes->NinetyPercentSample(curr - minRtt, curr - 0.001); */
    /*    float prevQmax = qSizes->PercentileSample(curr - minRtt, curr - 0.001, 80); */
    float currSample = qSizes->LookupSample(curr);
    float oldSample = qSizes->AverageSample(curr - 2*minRtt, curr - minRtt);
    if (compTime > curr - minRtt) {
	if (compQmax > currSample)
	    return;
    } else
	compQmax = minRtt;
    if (currSample < minRtt)
	return;

#if 1
    if ((oldSample - minRtt) > 0.1*(currSample - minRtt))
	factor = 0.1;
    else
	factor = 1;
#if 1
    if (startTime > curr - 6*minRtt)
	factor = 0.0;
#endif
#endif
    float newRate = uniformRate - uniformRate*((currSample - compQmax)/minRtt)*factor;
    if (newRate < 0.001) newRate = 0.001;
#ifdef DEBUG
    printf("%f: flow %d underlying rate %f --> %f\n", curr, flowNo, uniformRate, newRate);
#endif
    UpdateUnderlyingRate(newRate, curr);
    compQmax = currSample;
    compTime = curr;
}

void TscpSender::HandleMessage(Message *m, float curr)
{
    int recvdp = board->Receivedp(m->seqno);
    board->RecordAck(m->seqno, m->startTime);
    if (board->UnackedBytes() == 0 && payload == bytesSent) {
	isActive = 0;
	FinishFlow(curr);
	return;
    }
    
    if (m->seqno == 0) {
	rttEstimate = curr - m->startTime;
	/* initialize rttVarEstimate to something high */
	rttVarEstimate = (curr - m->startTime)/2;
	minRtt = rttEstimate;
	maxRtt = minRtt;
	Event *nextEv = new Event(curr + minRtt, this, RTT_TIMER, NULL, curr);
	net->AddEvent(nextEv);
    } else {
	float newRtt = curr - m->startTime;
	float newRttVar = fabs(newRtt - rttEstimate);

	rttEstimate = 0.875*rttEstimate + 0.125*newRtt;
	rttVarEstimate = 0.75*rttVarEstimate + 0.25*newRttVar;
	
	if (!recvdp) {
	    int msgSize = board->MsgSize(m->seqno);
	    if (msgSize == packetSize) {
		if (newRtt < minRtt)
		    minRtt = newRtt;
	    }
	}
	if (newRtt > maxRtt)
	    maxRtt = newRtt;
    }
    qSizes->RecordSample(curr, (curr - m->startTime));
    UCompensate(curr);

#ifdef DEBUG
    printf("%f: message %d:%d ack received, rtt: %f\n", curr, flowNo, m->seqno, curr - m->startTime);
#endif
    
    if (m->seqno == 0) {
	Agent *s = conn->agents[0];
	Agent *d = conn->agents[conn->numAgents - 2];
	float minrttEst;
	float bwest = net->LookupHistory(s, d, &minrttEst);

	if (bwest > 0) {
	    minRtt = (minRtt > minrttEst ? minrttEst : minRtt);
	    if (bwest/3 > (packetSize + protoOvhd)/minRtt) {
		float rate = (packetSize + protoOvhd)/minRtt;
		int phase;
		for (phase=0; ;phase++) {
		    if (rate > bwest/3)
			break;
		    rate = 2*rate;
		}
		burstPhase = phase;
	    }
	}
	StartNewBurst(curr);
	
    } else {

	if (burstMsgsSent > 0) {

	    for (int i=0; i<burstMsgsSent; i++) {
		if (burstMsgSeqno[i] == m->seqno) {
		    burstMsgRtt[i] = curr - m->startTime;
		    burstMsgsRcvd++;
		    break;
		}
	    }
	    
	    if (burstMsgsRcvd == 5) {
		ProcessBurstResults(curr);
		StartNewBurst(curr);
	    }
	}
    }

    int reordSeq = board->CheckRetransmit();
    if (reordSeq != -1 && !InDropQueue(reordSeq)) {
	RegisterDrop(reordSeq);
	RestartBurst(reordSeq, curr);
    }
    delete m;
}

TscpGlobalStats::TscpGlobalStats(Network *n, int maxflows, float bneck)
{
    net = n;
    maxFlows = maxflows;
    numActiveFlows = 0;
    flows = new (TscpSender *)[maxFlows];
    for (int i=0; i<maxFlows; i++)
	flows[i] = NULL;
    cumURate = cumBRate = 0;
    numBursting = 0;
    bottleneck = bneck;
    sprobeBWBins = new int [2*(int)bneck/50];
    fprobeBWBins = new int [2*(int)bneck/50];
    fnprobeBWBins = new int [2*(int)bneck/50];
    fpprobeBWBins = new int [2*(int)bneck/50];
    for (int j=0; j<2*bneck/50; j++)
	sprobeBWBins[j] = fprobeBWBins[j] = fnprobeBWBins[j] = fpprobeBWBins[j] = 0;

    nextSample = 0;
    nextBSample = 0;
    maxSamples = 10000;
    recentUrates = new float[maxSamples];
    recentTimes = new float[maxSamples];
    recentBrates = new float[maxSamples];
    recentBTimes = new float[maxSamples];
}

void TscpGlobalStats::StartFlowStats(TscpSender *s)
{
    int id = s->flowNo;
    assert(id < maxFlows);
    assert(flows[id] == NULL);
    flows[id] = s;
    cumURate += s->uniformRate;
    numActiveFlows++;
}

void TscpGlobalStats::StopFlowStats(TscpSender *s)
{
    int id = s->flowNo;
    assert(id < maxFlows);
    assert(flows[id] != NULL);
    flows[id] = NULL;
    cumURate -= s->uniformRate;
    numActiveFlows--;
}

void TscpGlobalStats::UpdateURateStats(float oldRate, float newRate, float curr)
{
    cumURate += (newRate - oldRate);
    recentUrates[nextSample] = cumURate;
    recentTimes[nextSample] = curr;
    nextSample = (nextSample + 1) % maxSamples;
}

void TscpGlobalStats::CleanStats()
{
    burstSuccesses = burstFailures = falsePositives = falseNegatives = 0;
    burstPacketloss = 0;
}

void TscpGlobalStats::StartBurstStats(float curr, float rate)
{
    numBursting++;
    cumBRate += rate;
    recentBrates[nextBSample] = cumBRate;
    recentBTimes[nextBSample] = curr;
    nextBSample = (nextBSample + 1) % maxSamples;
    PrintCumStats();
}

void TscpGlobalStats::StopBurstStats(float curr, float rate)
{
    numBursting--;
    cumBRate -= rate;
    recentBrates[nextBSample] = cumBRate;
    recentBTimes[nextBSample] = curr;
    nextBSample = (nextBSample + 1) % maxSamples;
    PrintCumStats();    
}

void TscpGlobalStats::RecordBurstSuccess(float rate, int succ, int cancel, float bursttime)
{
    if (!cancel) {
	if (succ)
	    sprobeBWBins[(int)rate/50]++;
	else
	    fprobeBWBins[(int)rate/50]++;

	if (succ)
	    burstSuccesses++;
	else
	    burstFailures++;

	int i = (nextSample - 1 + maxSamples) % maxSamples;
	while (recentTimes[i] > bursttime)
	    i = (i - 1 + maxSamples) % maxSamples;
	float cum = recentUrates[i];

	i = (nextBSample - 1 + maxSamples) % maxSamples;
	while (recentBTimes[i] > bursttime)
	    i = (i - 1 + maxSamples) % maxSamples;
	float bcum = recentBrates[i];
	
	if (((cum + bcum + rate) < 0.9*bottleneck) && !succ) {
	    falseNegatives++;
	    fnprobeBWBins[(int)rate/50]++;
	}
	else if (((cum + bcum + rate) > 1.1*bottleneck) && succ) {
	    falsePositives++;
	    fpprobeBWBins[(int)rate/50]++;
	}
    }
}
    

void TscpGlobalStats::PrintCumStats()
{
#ifdef DEBUG
    printf("%f: %d active flows, %f cum-u-rate, %d bursting, %f cum-burst-rate\n",
	   net->GetTime(), numActiveFlows, cumURate, numBursting, cumBRate);
#endif
}

void TscpGlobalStats::PrintStats()
{
    
#if 0
    printf("Bursts: %d (success: %d fail: %d +pos: %d +neg: %d, loss %d\n",
	   burstSuccesses + burstFailures, burstSuccesses, burstFailures,
	   falsePositives, falseNegatives, burstPacketloss);
    for (int j=0; j<2*bottleneck/50; j++)
	printf("%d %d %d %d %d\n", j*50+25, sprobeBWBins[j], fprobeBWBins[j],
	       fnprobeBWBins[j], fpprobeBWBins[j]);
#endif
}

void TscpGlobalStats::StopAllFlows(float curr)
{
#if 0
    for (int i=0; i<maxFlows; i++)
	if (flows[i])
	    flows[i]->FinishFlow(curr);
#endif
}

void TscpGlobalStats::PrintAllFlows(float curr)
{
    for (int i=0; i<maxFlows; i++)
	if (flows[i])
	    printf("%d %f %f %f\n", i, flows[i]->startTime,
		   flows[i]->uniformRate, flows[i]->minRtt);
}
