#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include "tscp.h"

static int uniqueId = 0;

Message *MakeMessage(int flowNo, int type, int size,
		     Connection *conn, int seqno,
		     int fin, float time)
{
    Message *m = new Message;
    m->flowNo = flowNo;
    m->msgType = type;
    m->uniqueId = uniqueId++;
    m->size = size;
    m->conn = conn;
    m->seqno = seqno;
    m->startTime = time;
    m->midTime = 0;
    m->endTime = 0;
    m->bwshare = 100000;
    m->qlen = 0;
    m->ack2 = -1;
    m->fin = fin;
    m->burstMsg = 0;
    return m;
}

Connection *MakeConnection(int num, Agent **agents)
{
    Connection *c = new Connection;
    c->numAgents = num;
    c->agents = new (Agent *)[num];
    for (int i=0; i<num; i++)
	c->agents[i] = agents[i];
    return c;
}

static int numSamples = 0;
static float cumResp = 0;
static float cumResp2 = 0;
static int bytesSent = 0;
static int numRexmits = 0;
static int numPackets = 0;
static int numUniquePackets = 0;

void TrackResponseTime(float t, int bytes, int numRex, int numP, int numUP)
{
    /*    printf("response: %f\n", t); */
    cumResp += t;
    cumResp2 += t*t;
    bytesSent += bytes;
    numRexmits += numRex;
    numPackets += numP;
    numUniquePackets += numUP;
    numSamples++;
}

void EnableStats()
{
    cumResp = 0;
    cumResp2 = 0;
    bytesSent = 0;
    numRexmits = 0;
    numSamples = 0;
}

extern float total;
extern float totalSamples;
void PrintCompressedStats(float period)
{
    printf("%f %f %f ", (1000*total)/period, cumResp/numSamples,
	   (float)sqrt(cumResp2/numSamples - (cumResp/numSamples)*(cumResp/numSamples)));
}

void PrintStats(float period)
{
    printf("Average response time of %d flows is %f, std-dev is %f\n",
	   numSamples, cumResp/numSamples,
	   (float)sqrt(cumResp2/numSamples - (cumResp/numSamples)*(cumResp/numSamples)));
    /*    PrintCompressedStats(period); */
}

#include <math.h>

float ThinkTimeDist(float mean)
{
#if 0
    float u = ((float)(random() % 10000))/10000.0;
    return mean*(-log(1-u));
#endif
    return mean*(0.7 + 0.6*((random() % 100)/100.0));
}

float ExpTimeDist(float mean)
{
#if 0
    float u = ((float)(random() % 10000))/10000.0;
    return mean*(-log(1-u));
#else
    return mean*(0.7 + 0.6*((random() % 100)/100.0));
#endif
}
