#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "tscp.h"

/*
   workload specification file starts with information
   regarding link characteristics.  then a tcp flow
   is specified by giving the sequence of links traversed,
   the number of bytes to be sent, and the start time
   of the flow.
*/

void ReadWorkload(Network *n, char *fname)
{
    FILE *fp = fopen(fname, "r");
#define MAX_LINE_SIZE	200
    char line[MAX_LINE_SIZE];
    char *res;
    Agent *linkAgents[200];
    Agent *flowAgents[200];
    TscpGlobalStats *tscpStats = new TscpGlobalStats(n, 1200, 5000);

    res = fgets(line, MAX_LINE_SIZE, fp);
    while (res) {
	char *s = line;
	char itemType[20];
	int pos, items;

	items = sscanf(s, "%s %n", itemType, &pos);
	if (items == 1) {
	    if (strcmp(itemType, "fifolink") == 0) {
		int linkId, queueSize;
		float bw, delay;
		/* get link characteristics */
		/* format: "fifolink linkId queueSize bw-in-kB/s delay-in-ms pktSize" */
		s += pos;
		sscanf(s, "%d %d %f %f %d", &linkId, &queueSize, &bw, &delay);
		assert(linkId < 100);
		linkAgents[2*linkId] = new FifoLink(n, queueSize*1250, bw, delay, 0);
		linkAgents[2*linkId + 1] = new FifoLink(n, queueSize*1250, bw, delay, 0);
	    } else if (strcmp(itemType, "fqlink") == 0) {
		int linkId, queueSize;
		float bw, delay;
		int maxflows, maxflowid;
		int pktSize;
		/* get link characteristics */
		/* format: "fqlink linkId queueSize pktSize bw-in-kB/s delay-in-ms maxflows maxflowid" */
		s += pos;
		assert(sscanf(s, "%d %d %d %f %f %d %d", &linkId, &queueSize, &pktSize,
			      &bw, &delay, &maxflows, &maxflowid) == 7);
		assert(linkId < 100);
		linkAgents[2*linkId] = new FQLink(n, queueSize*pktSize, pktSize, bw, delay, maxflows, maxflowid, 0);
		linkAgents[2*linkId + 1] = new FQLink(n, queueSize*pktSize, pktSize, bw, delay, maxflows, maxflowid, 0);
	    } else if ((strcmp(itemType, "pcpflow") == 0) ||
		       (strcmp(itemType, "tcpflow") == 0) ||
		       (strcmp(itemType, "fqflow") == 0)) {
		/* format: "pcpflow starttime bytes pktSize linkid1 linkid2 ... */
		int links[20];
		Agent *agents1[20];
		Agent *agents2[20];
		int numlinks = 0;
		float starttime;
		int numBytes;
		int pktSize;

		s += pos;
		items = sscanf(s, "%f %d %d %n", &starttime, &numBytes, &pktSize, &pos);
		assert(items == 3);
		s += pos;
		while (s) {
		    items = sscanf(s, "%d %n", &links[numlinks], &pos);
		    if (items == 1) {
			numlinks++;
			assert(numlinks < 20);
			s += pos;
		    } else
			break;
		}
		for (int i=0; i<numlinks; i++) {
		    agents1[i] = linkAgents[2*links[i]];
		    agents2[numlinks-1-i] = linkAgents[2*links[i] + 1];
		}
		GenericReceiver *r = new GenericReceiver(n);
		agents1[numlinks] = r;
		Connection *c1 = MakeConnection(numlinks+1, agents1);
		if (strcmp(itemType, "pcpflow") == 0) {
		    TscpSender *s = new TscpSender(n, numBytes, pktSize, -1, tscpStats);
		    agents2[numlinks] = s;
		    Connection *c2 = MakeConnection(numlinks+1, agents2);
		    s->SetConnection(c1);
		    r->SetConnection(c2);
		    s->Start(starttime);
		} else if (strcmp(itemType, "tcpflow") == 0) {
		    TcpSender *s = new TcpSender(n, numBytes, pktSize, -1, NULL);
		    agents2[numlinks] = s;
		    Connection *c2 = MakeConnection(numlinks+1, agents2);
		    s->SetConnection(c1);
		    r->SetConnection(c2);
		    s->Start(starttime);
		} else if (strcmp(itemType, "fqflow") == 0) {
		    FQSender *s = new FQSender(n, numBytes, pktSize, -1);
		    agents2[numlinks] = s;
		    Connection *c2 = MakeConnection(numlinks+1, agents2);
		    s->SetConnection(c1);
		    r->SetConnection(c2);
		    s->Start(starttime);
		}
	    }
	}
	res = fgets(line, MAX_LINE_SIZE, fp);
    }
    fclose(fp);
}
